require('./global/index');
const fs = require('fs');
const ora = require('ora');
const chalk = require('chalk');
// const webpack = require('webpack');
const path = require('path');
const qrcode = require('qrcode-terminal');
// const zipDir = require('zip-directory');
const superagent = require('superagent');
const extract = require('extract-zip');

const { launchStaticServer } = require('./pipeline/launch-server');
const { Compiler } = require('./pipeline/webpack-build');
const { zip } = require('./pipeline/zip');
const { getDebugUrl } = require('./pipeline/debug-tool');

const renderConfig = require('./webpack.render');
// const serviceConfig = require('./webpack.service');
const serviceConfigBase = require('./splitChunksConfig/webpack.service.base');
const serviceConfig = require('./splitChunksConfig/webpack.service');
const getIpAddress = require('./getIpAddress');

const ROOT = process.cwd();
const SRC = path.join(ROOT, 'src');
const DIST = path.join(ROOT, 'dist');
let wsInstance = null;

console.log('[biliSapp - info] 启动编译')

if (!fs.existsSync(path.join(ROOT, 'zip'))) {
    fs.mkdirSync(path.join(ROOT, 'zip'));
}

// zip 目录建立 static host，用于给 app 下载 zip 包
if (!process.env.TEST_BUILDER && process.argv.indexOf('no-serve') === -1) {
    launchStaticServer(path.join(ROOT, 'zip'));
}

// 强制校验 mid。
let midStr = '';

process.argv.some(arg => {
    if (arg.indexOf('mid') === 0) {
        midStr = arg;
        return true;
    }
});

if (midStr === '') {
    console.log('[biliSapp - error]:请使用 mid 命令行进行构建，命令：npm run build mid=xxx \r\n');
    process.exit();
}
console.log('[biliSapp - info] 校验 mid 成功！');

// const spinner = ora('building for pages...');
// spinner.start();

if (!fs.existsSync(path.join(ROOT, 'src/service.js'))) {
    fs.writeFileSync(path.join(ROOT, 'src/service.js'), '', 'utf-8');
}

// 检查用户是否忽略工具判断
const _ignore_tool_check = process.argv.indexOf('ignore_tool_check') > -1;

if (!_ignore_tool_check) {
    console.log('[biliSapp - info] 开始检查编译工具版本');

    // package.json 中的 small-app 的版本号
    let packageSapp = ''
    // 本地安装的small-app的版本号（包含：安装、没有安装）
    let localSappVersion = ''
    // 远端的small-app的版本号最新
    let serveSappVersion = ''
    // 远端的small-app的所有tag号
    let serveSappAllTags = ''
    // 本地安装 bili-small-app 的信息
    let localSapp = {};

    // 检查本地 small-app版本号
    const packageJson = fs.readFileSync(path.join(ROOT, 'package.json'), 'utf8');

    try {
        const _packageJson = JSON.parse(packageJson);
        packageSapp = _packageJson.dependencies['bili-small-app'];
    } catch (e) {
        console.log('[biliSapp - warn]:检测到当前小程序未安装「 bili-small-app 」。')
    }

    console.log('[biliSapp - info] 本地用户指定编译工具版本：', packageSapp);

    const exec = require('child_process').spawnSync;
    // 检查远端 small-app版本号
    try {
        const serveJSON = exec('npm',['view', 'bili-small-app', '--json']).stdout.toString();
        serveSappVersion = JSON.parse(serveJSON).version;
        serveSappAllTags = JSON.parse(serveJSON)['dist-tags'];
    } catch (e) {
        console.log('[biliSapp - warn]:', '检查「 bili-small-app 」最新版本失败，请在控制台检查「 npm view bili-small-app 」命令是否可用')
    }

    try {
        localSapp = JSON.parse(exec('npm',['list', 'bili-small-app', '--json']).stdout.toString());
        localSappVersion = localSapp.dependencies['bili-small-app'].version;
    } catch (e) {
        console.log('[biliSapp - warn]:', '检查「 bili-small-app 」本地安装版本失败，请在控制台检查「 npm list bili-small-app 」命令是否可用')
    }

    const isTagBySapp = !!serveSappAllTags[packageSapp]
    const _isTagByLocal = serveSappAllTags[packageSapp] !== localSappVersion

    console.log('[biliSapp - info] 远端编译工具最新版本：', isTagBySapp? serveSappAllTags[packageSapp] : serveSappVersion);
    console.log('[biliSapp - info] 实际安装编译工具版本：', localSappVersion || '未安装');

    if (!isTagBySapp) {
        console.log('[biliSapp - info] 编译工具更新中...');

        const _execResult = exec('npm',['i', 'bili-small-app@latest'], {
            cwd: path.join(ROOT, './')
        })
        if (_execResult.status === 0) {
            console.log('[biliSapp - info] 编译工具更新成功！');
        } else {
            console.log(`[biliSapp - info] 编译工具更新失败，请在控制台确认命令「npm i bili-small-app@latest」命令是否可用`);
        }

        console.log('[biliSapp - info] 编译工具更新成功！');

        let temp = JSON.parse(packageJson);
        temp.dependencies['bili-small-app'] = 'latest';

        fs.writeFileSync(path.join(ROOT, 'package.json'), JSON.stringify(temp, undefined, '\t'),'utf8');

    } else {
        if (_isTagByLocal) {
            console.log('[biliSapp - info] 编译工具更新中...');
            const _execResult = exec('npm',['i', `bili-small-app@${packageSapp}`], {
                cwd: path.join(ROOT, './')
            })
            if (_execResult.status === 0) {
                console.log('[biliSapp - info] 编译工具更新成功！');
            } else {
                console.log(`[biliSapp - info] 编译工具更新失败，请在控制台确认命令「npm i bili-small-app@${packageSapp}」`);
            }

            let temp = JSON.parse(packageJson);
            temp.dependencies['bili-small-app'] = packageSapp;

            fs.writeFileSync(path.join(ROOT, 'package.json'), JSON.stringify(temp, undefined, '\t'),'utf8');
        }
        if (packageSapp !== 'latest') {
            console.log('[biliSapp - warn] 当前是开发版，上线前请替换「bili-small-app」的版本为「latest」');
        }
    }
}

let buildType = 'sapp';
let mode = 'build';
let base = null;
process.argv.forEach(arg => {
    if (arg.indexOf('--watch') === 0 || arg.indexOf('-w') === 0) {
        mode = 'watch';
    } else if (arg.indexOf('--widget') === 0) {
        buildType = 'widget';
    } else if (arg.indexOf('base') === 0) {
        base = arg.split('=')[1];
    }
});

console.log('[biliSapp - info] 开始编译小程序...');

// const compiler = new Compiler([renderConfig(ROOT), serviceConfig(ROOT, SRC)]);
const compiler = new Compiler([renderConfig(ROOT, 'development'), serviceConfigBase(ROOT, SRC, 'development'), serviceConfig(ROOT, SRC, 'development')]);

if (mode === 'build') {
    compiler.build(false).then(() => {
        compilerDone();
    });
} else if (mode === 'watch') {
    compiler.watch(false, compilerDone);
}

async function compilerDone() {
    // fs.unlink(path.join(SRC, 'service.js'), () => {});
    console.log('[biliSapp - info] 编译小程序完成');

    // 兜底，分包如果没打出common.service.js，则手动写一个空的，因为客户端逻辑是加载到common.service.js才进行分包逻辑
    if (!fs.existsSync(path.join(DIST, './common.service.js'))) {
        fs.writeFileSync(path.join(DIST, './common.service.js'), '"";', 'utf-8');
    }

    // spinner.stop();

    const zipName = new Date().toJSON().replace(/[-:.TZ]/g, '') + '.zip';
    if (!fs.existsSync(path.join(ROOT, 'zip'))) {
        fs.mkdirSync(path.join(ROOT, 'zip'));
    }
    const zipPath = path.join(ROOT, 'zip', zipName);
    let zipLatestName = 'latest.zip';

    if (mode === 'watch') {
        zipLatestName = 'watch.zip';
    }
    const latestZipPath = path.join(ROOT, 'zip', zipLatestName);

    const zipFiles = [latestZipPath];
    if (mode === 'build') {
        zipFiles.push(zipPath);
    }

    if (base) {
        if (!path.isAbsolute(base)) {
            base = path.resolve(process.cwd(), base);
        }
        const targetDist = path.resolve(process.cwd(), 'dist/');

        if (!fs.existsSync(base)) {
            console.error('base 路径不存在，请检查路径是否正确');
            process.exit();
        }

        try {
            await extract(base, { dir: targetDist });
        } catch (e) {
            console.error(e);

            process.exit();
        }
    }

    zip(DIST, ...zipFiles).then(async() => {
        const ip = getIpAddress();
        const sapprc = JSON.parse(fs.readFileSync(path.join(ROOT, '.sapprc'), 'utf8'));
        let storageId = '';
        // let debugURL = `https://miniapp.bilibili.com/applet/debug/${sapprc.appId}_${sapprc.vAppId}?url=${encodeURIComponent(`http://${ip}:3000/${zipName}`)}`;
        // let debugURL = `https://miniapp.bilibili.com/applet/debug/${sapprc.appId}_${sapprc.vAppId}?url=${encodeURIComponent(`http://${ip}:3000/${zipLatestName}`)}`;
        let debugURL = getDebugUrl({
            type: buildType,
            ip,
            zipName: zipLatestName,
            sapprc,
            embedBase: !!base
        });

        if (midStr) {
            // mid=xxx
            let mid = midStr.slice(4);

            try {
                let uploadUrl = '';

                let upZipResponse = await superagent
                    .post('http://mall-mng.bilibili.co/mall-mng/picture/image')
                    .attach('file', latestZipPath, zipName);
                if (upZipResponse.status === 200 && upZipResponse.body.code === 0 && upZipResponse.body.data) {
                    uploadUrl = 'http:' + upZipResponse.body.data;
                } else {
                    console.log('本地构建失败，请重试');
                    process.exit();
                    return;
                }

                // 本地调试校验参数判断
                let _tempParam = {appId: sapprc.appId, url: uploadUrl, isValidUrl: 0 }
                if (global._bili_cache_data.isInternal) {
                    _tempParam.staff = mid;
                } else {
                    _tempParam.mid = mid;
                }

                let upResponse = await superagent
                    .post('http://miniapp.bilibili.com/api/storage/dev_storage/upload')
                    .send(_tempParam)
                    .set('Accept', 'application/json');
                if (upResponse.status === 200 && upResponse.body.code === 0) {
                    storageId = upResponse.body.data;
                    debugURL = getDebugUrl({
                        type: buildType,
                        ip,
                        zipName: zipLatestName,
                        sapprc,
                        storageId
                    });
                } else {
                    console.log('本地构建失败：', upResponse.body.message);
                    process.exit();
                    return;
                }
            } catch (error) {
                console.log('本地构建失败，请重试');
                process.exit();
                return;
            }
        }

        // console.log(chalk.cyan('  Build service complete.\n'));
        if (process.env.TEST_BUILDER || process.argv.indexOf('no-serve') !== -1) {
            return;
        }

        let devtoolPort;
        // 调试工具静态端口一般从10022开始
        // 发现window版本扫描接口较慢，这里就不做过多的扫描了
        for (let i = 10022; i < 10028; i++) {
            try {
                const response = await superagent.post(`http://localhost:${i}/devtool-scan`).send({});
                if (response.status === 200 && response.text === 'small-app-devtool-application listened...') {
                    devtoolPort = i;
                    break;
                }
            } catch (error) {
                // console.error(error);
            }
        }
        if (devtoolPort) {
            await superagent.post(`http://localhost:${devtoolPort}/devtool-debug`).send({ debugURL });
            console.log(chalk.cyan('检测到调试工具正在运行，可以使用调试工具进行调试啦'));
        } else {
            console.log('[biliSapp - info] 生成二维码');
            qrcode.generate(debugURL, {small: true});
            console.log(debugURL);
            console.log(chalk.cyan('真机调试服务正在运行，可以用手机扫码调试构建出来的安装包啦'));
        }
        console.log(chalk.cyan('若要重新构建需退出（ctrl+c），再执行一次 npm run build'));

        wsInstance && wsInstance.send('hotreload');
    });
}
